#!/usr/bin/env python2.7


import numpy as np
import rospy
from geometry_msgs.msg import Point
from geometry_msgs.msg import PoseStamped
from geometry_msgs.msg import TwistStamped
from sensor_msgs.msg import LaserScan
from std_msgs.msg import Header

def velocity_callback(data):
    global velocity
    velocity = (data.twist.linear.x,data.twist.linear.y)
    

def pose_callback(data):
    global pose
    pose = data.pose.position

def intersection(line1, line2):
    #get x of intercept from  x = b2-b1/m1-m2
    x = (line2[1]-line1[1])/(line1[0]-line2[0])
    #get y of intercept from y = mx+b (for either line
    y = x*line1[0]+line1[1]
    return np.array([x,y])

def distance_to_line(line1,vector1, center1, line2, bound1, bound2):
    intersect = intersection(line1,line2)
    #if the intersection point is outside where the line actually exists, return inifinity
    if not (intersect[0]>bound1 and intersect[0]<bound2):
        return float('inf')
    recentered_intersect = intersect-center1
    distance = recentered_intersect[0]/vector1[0]
    if distance-recentered_intersect[1]/vector1[1]>0.01:
        raise Exception('distance not the same')
    #if the distance is negative, the point is behind, so we return infinity to show nothing ahead
    if distance<0:
        return float('NaN')
    #if the drone gets too close of the wall, say the drone crashed
    if distance<0.5:
        raise Exception('crashed')

    return distance

def calculate_distance_array():
    #this method calculates the distances to walls in 61 different directions
    #with the center (31st) direction being straight ahead, and each direction being
    #seperated by 3 degrees
    #first, we initialize the array
    distances = np.zeros(61)
    #and the drones current pose
    center = np.array([pose.x,pose.y])
    #next we calculate the first angle, which is the angle of the first item in
    #the directions array. This is calculated as the direction the drone is 
    #travelling minus 30 times the angle between each element of the array
    first_angle = np.arctan(velocity[1]/velocity[0])-30*unit_degree
    if velocity[0]<0:
        first_angle+=np.pi
    #next we populate the array
    for x in range(61):
        #for each element, we first calculate the angle that the distance will
        #be for
        angle = first_angle+x*unit_degree
        #next we calculate a unit vector in that direction
        vector = np.array([np.cos(angle),np.sin(angle)])
        #next we calculate the equation for the line.
        line = [0,0]
        #the m value is calculated as m = delta y/delta x, with delta x and y 
        #being from the vector we just calculated
        line[0] = vector[1]/vector[0]
        #the b value is calculated as b = y-mx. Here we use pose of the drone
        #for x and y values
        line[1] = center[1]-line[0]*center[0]
        #next for each wall, we calculate the distance to the wall, and we take
        #the smallest distance
        distance = float('inf')
        for line_boundary_set in wall_lines:
            temp_distance = distance_to_line(line,vector, center,line_boundary_set[0],line_boundary_set[1],line_boundary_set[2])
            if temp_distance<distance:
                distance = temp_distance
        distances[x] = distance
    #after we calculate the distances for each wall, we return them
    return distances

def distance_publisher():
    #first we set up our publishers
    pub = rospy.Publisher('obstacle_distances',LaserScan, queue_size = 10)
    rospy.init_node('obstacle_publisher',anonymous=True)
    rospy.Subscriber('mavros/local_position/velocity_body', TwistStamped, velocity_callback)
    rospy.Subscriber('mavros/local_position/pose', PoseStamped, pose_callback)
    #we set up the publishing rate
    rate = rospy.Rate(10)
    #we sleep to make sure that we get our first pose/velocity measurements
    #before we try and determine distances
    rospy.sleep(10)
    #finally, we start publishing
    while not rospy.is_shutdown():
        #first we calculate the distances
        distances = calculate_distance_array()
        #then we set up the message, and publish it
        to_publish = LaserScan()
        to_publish.angle_min = -30*unit_degree
        to_publish.angle_max = 30*unit_degree
        to_publish.angle_increment = unit_degree
        to_publish.time_increment = 0
        to_publish.scan_time = 0.1 
        to_publish.range_min = 0
        to_publish.range_max = float('inf')
        to_publish.ranges = distances
        pub.publish(to_publish)
        rate.sleep

if __name__ == '__main__':
    #unit_degree defines the amount the change in angle between scans
    unit_degree = 3*np.pi/180
    pose = None
    #wall_lines is a variable that defines all of the lines that make up 
    #obstacles, as well as their boundaries wall lines is a list of tuples. 
    #Each defines a wall, and is written in the form
    #([m,b],lower bound for x, upper bound for x)
    #where the line is defined by y=mx+b, but only exists when x is between
    #the lower and upper bounds.
    wall_lines = [ ([-1,150],-100,100),([1,350],-100,100)]

    try:
        distance_publisher()
    except rospy.ROSInterruptException:
        pass



    











